# although screws up in twinview. hmmm..
#' get screen resolution of a monitor (may not work)
#' @param quietly whether to fail quietly if the screen resolution can not
#' be found, or throw an error.
#' @return c(horizontal,vertical) screen resolution. NULL if not found.
#' @note **NOT GUARANTEED TO WORK**: calls \code{xdpyinfo} on linux and
#' \code{wmic} on windows, which isn't always installed. Who knows what
#' happens on Mac.
#' @family units
#' @export
getScreenResolution <- function(quietly=FALSE)
{
    # unix...
    if ( !is.null((PROG <- getProgram('xdpyinfo'))) ) {
        tmp <- system(sprintf("%s | grep dimensions",PROG[1]), intern=TRUE)
        tmp <- strsplit(tmp,"[ ]")[[1]]
        tmp <- tmp[grep('\\b[0-9]{3,5} ?[Xx] ?[0-9]{3,5}\\b',tmp,perl=TRUE)][1]
        szs <- as.integer(strsplit(tmp,'[Xx]')[[1]])  # size of screen
    } else if ( !is.null((PROG <- getProgram(c('wmic','wmic.exe')))) ) {
    # windows...
        scr_width <- system(sprintf("%s desktopmonitor get screenwidth",
                                     PROG[1]), intern=TRUE)
        # [1] "ScreenWidth  \r" "1440         \r" "\r"
        scr_height <- system(sprintf("%s desktopmonitor get screenheight",
                                     PROG[1]), intern=TRUE)
        szs <- as.numeric(c(
                       scr_width[-c(1, length(scr_width))],
                         scr_height[-c(1, length(scr_height))]
        ))
    } else {
        if ( quietly )
            szs <- NULL
        else
            stop("getScreenResolution: could not get screen size")
    }

    return(szs)
}

#' get the pixels per inch of a screen (may not work!)
#' @param quietly whether to fail quietly if the screen resolution can not
#' be found, or throw an error.
#' @return c(horizontal,vertical) PPI (possibly just one value though on
#' windows?)
#' @note **NOT GUARANTEED TO WORK**: calls \code{xdpyinfo} on linux and
#' \code{wmic} on windows, which isn't always installed. Who knows what
#' happens on Mac.
getPPI <- function(quietly=FALSE)
{
    # unix... xdpyinfo
    if ( !is.null((PROG <- getProgram('xdpyinfo'))) ) {
        tmp <- system(sprintf("%s | grep resolution",PROG[1]), intern=TRUE)
        tmp <- strsplit(tmp,"[ ]")[[1]]
        tmp <- tmp[grep('\\b[0-9]{2,4} ?[Xx] ?[0-9]{2,4}\\b',tmp,perl=TRUE)][1]
        ppi <- as.integer(strsplit(tmp,'[Xx]')[[1]])  # size of screen
    } else if ( !is.null((PROG <- getProgram(c('wmic','wmic.exe')))) ) {
    # windows...
    # unsure if this works
        ppi_width <- system(sprintf("%s desktopmonitor get pixelsperxlogicalinch",
                                     PROG[1]), intern=TRUE)
        # [1] "ScreenWidth  \r" "1440         \r" "\r"
        ppi_height <- system(sprintf("%s desktopmonitor get pixelsperylogicalinch",
                                     PROG[1]), intern=TRUE)
        ppi <- as.numeric(c(
                       ppi_width[-c(1, length(ppi_width))],
                         ppi_height[-c(1, length(ppi_height))]
        ))
    } else {
        if ( quietly )
            ppi <- NULL
        else
            stop("getPPI: could not get DPI/PPI")
    }

    return(ppi)
}

#' Attempt to convert between npc, inches, and pixels on the screen.
#' @param value value to convert
#' @param from unit to convert from. See Details.
#' @param to unit to convert to. See Details.
#' @param type either 'width' or 'height'. Only of use for 'npc'
#' conversions -- whether you want to return a fraction of the height or of
#' the width.
#' @return the converted unit
#' @details
#' The available units are:
#' \describe{
#'     \item{inches}{Inches}
#'     \item{npc}{Normalised Parent Coordinates: the entire monitor has
#'     dimensions (1,1).}
#'     \item{pixels}{pixels}
#' }
#'
#' ** The conversions are rough only and not exact! Also, 'npc' conversions
#' may screw up with dual monitor!  **
#'
#' The conversion between pixels and inches uses
#' \code{imageUtilities.option('ppi')} (pixels per inch), which is default
#' 96.
#'
#' The conversion between pixels or inches and screen size uses the
#' function \code{\link{getScreenResolution}}, which doesn't
#' always work. If the code is unable to return the screen resolution, a
#' default value of
#' \code{imageUtilities.option('default.screen.resolution')} is used
#' (currently 1280 x 1024).
#'
#' @seealso \code{\link{imageUtilities.option}}
#' @examples
#' # convert 7 screen inches to screen pixels (approx).
#' convertScreenUnits(7,from='inches',to='pixels')
#' # find out what half the screen width is in inches
#' convertScreenUnits(.5,from='npc',to='inches',type='width')
#' # find out what fraction of the screen height 2 inches is.
#' convertScreenUnits(2,from='inches',to='npc',type='height')
#'
#' @family units
#' @export
convertScreenUnits <- function(value,from,to,type=c('width','height'))
{
    valid.units <- c('inches','npc','pixels')
    from <- match.arg(tolower(from),valid.units)
    to   <- match.arg(tolower(to),valid.units)

    if ( from == to )
        return(value)

    type <- match.arg(tolower(type),c('width','height'))
    if ( from == 'npc' || to == 'npc' ) {
        sz <- getScreenResolution(quietly=TRUE) # horiz, vert
        if ( is.null(sz) )
            sz <- imageUtilities.option('default.screen.resolution')
        sz <- sz[ifelse(type=='width',1,2)]
    }

    if ( from == 'inches' || to == 'inches' ) {
        ppi <- getPPI(quietly=TRUE)
        if ( is.null(ppi) )
            ppi <- imageUtilities.option('ppi')
        if ( length(ppi)>1 )
            ppi <- ppi[ifelse(type=='width',1,2)]
    }

    # note - could reduce some calculates by just calling
    # convertScreenUnits again, but I won't.
    if ( from == 'inches' ) {
        px <- value*ppi
        if ( to == 'pixels' )
            return(px)
        else if ( to == 'npc' )
            return(px/sz)
    } else if ( from == 'npc' ) {
        px <- value*sz
        if ( to == 'pixels' )
            return(px)
        else if ( to == 'inches' )
            return(px/ppi)
    } else if ( from == 'pixels' ) {
        if ( to == 'inches' )
            return( value/ppi )
        else if ( to == 'npc' )
            return( value/sz )
    }

    stop("Unrecognised to/from pair (contact developer!)")
    return(NULL)
}

#' returns the type of the given unit.
#' @param unit the input unit to get the type of
#' @param default.units what to return of the input has no unit type
#' @return the unit: either a grid unit type or \code{default.units} as a
#' string.
#' @family units
#' @export
#' @examples
#' library(grid)
#' unitType( unit(1,'inches') )           # 'inches'
#' unitType( unit(.5,'npc') )             # 'npc'
#' unitType( 1234 )                       # 'pixels'
#' unitType( 1234, default.units='fdsa' ) # 'fdsa'
unitType <- function(unit,default.units='pixels')
{
    unt <- attr(unit,'unit')
    if ( is.null(unt) )
        unt <- default.units
    return(unt)
}

# *** ARGH NON-Voir version ***
##################
# NEEDS GRID
##################
#' plots an image on an R device using the "grid" package.
#' @param img a heightxwidthxchannels *array* with the number of channels
#' @param max max intensity of image
#' @param dev.new whether to open up new device (sometimes may be nice not
#' to in order to draw on an existing device).
#' @param transpose boolean, whether to transpose image before drawing
#' (should not need to be used -- just for compatibility between Voir and
#' non-Voir versions of the package)
#' @param maxWidth the maximum width the image should be. Can be either a
#' number (in which case it will be interpreted as *pixels*), or a unit of
#' the form \code{unit(width,type)} where \code{type} is one of 'npc' or
#' 'inches' (see \code{\link{convertScreenUnits}}).
#' The *display* will be resized to approximately this size (although the
#' image displayed will not be subsampled in any way). However, if
#' parameter \code{resize} is \code{TRUE}, this changes the behaviour.
#' @param resize boolean. If this is set to TRUE, then instead of resizing
#' the *display* to \code{maxWidth}, the *image* will be subsampled to this
#' size (although the scale will still be as for the full-size image so
#' that \code{grid.locator()} still returns the right coordinates.
#' @param ... passed into \code{viewport} command used to draw the image.
#' @return handle to the viewport.
#' @note The coordinate system is set up such that the *middle* of the
#'       top-left pixel has coordinates (1,1) (ie 1-based).
#'       Y increases downwards and X increases left-to-right (ie, an i-j
#'       coordinate system).
#'       
#'       The middles of pixels have integer coordinates.
#' @family draw
#' @export
drawImage <- function(img,
                      max=imageUtilities.option('max'),
                      dev.new=TRUE,
                      maxWidth=unit(7,'inches'),
                      resize=FALSE,
                      transpose=imageUtilities.option('.using.Voir'),
                      ...) 
{
    library(grid)
    # @TODO: should I coerce to array here??
    sz <- imgetdims(img)

    if ( transpose )
        img <- imtranspose(img)

    # subsample if necessary???
    wid <- as.numeric(maxWidth)
    unt <- unitType(maxWidth,default.units='pixels')
    if ( resize ) {
        px <- convertScreenUnits(wid,from=unt,to='pixels')
        if ( (d<-sz[1]/px)>1 )
            img <- imsubsample(img,d,d,'every')
    }

    # this operates on images that are *arrays*
    if ( !is.array(img) )
        img <- im2array(img)

    if ( dev.new ) {
        # must work out new width in inches.
        base <- convertScreenUnits(wid,from=unt,to='inches')
        if ( is.null(base) ) base <- 7 # default.
        dev.new( width=base, height=sz[2]/sz[1]*base )
    }
    # NOTE: TO HAVE THE CENTRE OF THE PIXEL BEING (i,j)
    # AND TO HAVE THE CENTRE OF TOP-LEFT PIXEL BEING 1,1:
    vp <- viewport(xscale=c(.5,sz[1]+.5),yscale=c(sz[2]+.5,.5),...)
    # vp <- viewport(xscale=c(0,sz[2]-1),yscale=c(sz[1]-1,0))
    # pushViewport(vp)
    grid.raster(img/max,
                x=.5,y=.5,
    #           x=0,y=0,
                height=-sz[2],width=sz[1],
                default.units='native',just=c('left','top'),
                vp=vp,
                interp=FALSE)
    return(vp)
}

#' interactive interface for selecting points on an image (grid).
#' @param imgArray  the input image *array* to have the user click on.
#'                  Supply this, OR
#' @param vp        handle to an existing viewport where the image is
#'                  already plotted
#' @param numPoints the number of points to have the user select.
#'                  Default 4.
#' @param VERTEX_COLOUR R colour (string) specifying the colour to draw the
#'                      points in on the graphics window.
#'                      NULL to omit, default "yellow".
#' @param LINE_COLOUR   R colour (string) specifying the colour to draw
#'                      lines between the points (NULL to omit).
#'                      Default 'white'
#' @param ...       arguments passed in to \code{\link{drawImage}}   
#' @return a numPointsx2 array (X,Y) of points the user clicked on.
#'         The points are \bold{1-based} and the *middle* of the top-left
#'         pixel has coordinates (1,1).
#' @family draw
#' @export
pointPicker <- function(imgArray=NULL,vp=NULL,numPoints=4,VERTEX_COLOUR="yellow",LINE_COLOUR="white",...) 
{
    require(grid)
    if ( missing(imgArray) & missing(vp) )
        stop("pointPicker: at least one of imgArray and vp must be provided")
    if ( missing(vp) ) {
        vp <- drawImage(imgArray,...)
        sz <- imgetdims(imgArray)[c(2,1)]
    } else {
        sz <- c(abs(vp$yscale[1]-vp$yscale[2]),
                abs(vp$xscale[1]-vp$xscale[2]))
    }
    pushViewport(vp)
    pts <- array(NA,dim=c(numPoints,2),dimnames=list(NULL,c('X','Y')))
    npts <- 0
    ptsize <- min(sz[1:2])*.02
    while ( npts < numPoints ) {
        coords <- grid.locator()
        npts   <- npts + 1
        cat(sprintf("Point %i: %i,%i\n",
                    npts,round(coords$x),round(coords$y)))
        pts[npts,] <- c(coords$x,coords$y)

        # draw line from previous to current
        if ( !is.null(LINE_COLOUR)&&!is.na(LINE_COLOUR) && npts > 1 ) {
            grid.lines(x=pts[(npts-1):npts,'X'],y=pts[(npts-1):npts,'Y'],
                       default.units='native',
                       gp=gpar(col=LINE_COLOUR,lwd=2),
                       vp=vp)
        }

        # draw last line
        if ( !is.null(LINE_COLOUR)&&!is.na(LINE_COLOUR) && npts == 4 ) {
            grid.lines(x=pts[c(4,1),'X'],y=pts[c(4,1),'Y'],
                       default.units='native',
                       gp=gpar(col=LINE_COLOUR,lwd=2),
                       vp=vp)
        }

        # points
        if ( !is.null(VERTEX_COLOUR) && !is.na(VERTEX_COLOUR) )
        grid.points(x=coords$x,y=coords$y,size=unit(ptsize,"native"),
                    pch=19,gp=gpar(col=VERTEX_COLOUR),
                    vp=vp)
    }
    return(list(vp=vp,points=pts))
}

#' saves an image using the in-built png, bmp, tiff etc. devices
#'  and grid graphics.
#' @param img  the image to be saved
#' @param name the name to save the image as
#' @param FUN  one of \code{\link[grDevices]{jpeg}}, 
#'             \code{\link[grDevices]{png}},
#' @param max  the max intensity in the image (default 0-1 intensities as
#' opposed to 0-255, since this is a grid function).
#' @param switch.off boolean, whether to call \code{dev.off} after drawing
#' the image to close the device. It can be useful to have this FALSE if on
#' wishes to continue plotting after calling the function.
#' @param ...  get fed into \code{FUN}.
#' \code{\link[grDevices]{tiff}} or \code{\link[grDevices]{bmp}}.
#' @family drawing
#' @export
#' @keywords internal
imsave.grid <- function(img,name,FUN=tiff,
                        max=imageUtilities.option('max'),
                        switch.off=TRUE,...)
{
    if ( is.character(FUN) ) {
        fName <- FUN
        FUN   <- match.fun(fName)
    } else {
        fName <- as.character(substitute(FUN))
    }
    if ( !(fName %in% c('tiff','bmp','jpeg','png')) )
        stop('imsave.grid: "FUN" should be one of tiff, "bmp", jpeg, or png')

    sz <- dim(img)
    FUN(filename=name,width=sz[2],height=sz[1],...)
    vp <- drawImage(img,max=max,dev.new=FALSE)
    if ( switch.off )
        dev.off()
    return(vp)
}

#' overlays a single-channel binary image onto another.
#' @param img     the input image
#' @param ovl     the overlay image. 
#'                It will be coerced to binary.
#'                Where it is 1 or TRUE, the overlay will be shown on the
#'                input image.
#' @param transparency the transparency of the overlay, from 0 to 1.
#'                     0 corresponds to an opaque overlay, and 
#'                     1 corresponds to a transparent overlay.
#' @param col     The colour to draw the overlay in.
#'                Either a string of a colour recognised by R, or
#'                 a numeric vector \code{c(R,G,B)} with red/green/blue
#'                 values the desired colour, from 0 to 255.
#' @param max The maximum intensity of the input image (default 255 for
#' Voir, 1 for non-Voir).
#' @return an image with the overlay on top of it in the specified
#'         transparency and colour.
#'
#' @note This essentially performs the calcuation:
#' \deqn{ outI = transparency * colour * overlay + 
#'               (1-transparency) * image }
#' @family overlay
#' @keywords internal
#' @export
makeOverlay.grid <- function( img, ovl, transparency=0.5, 
                        max=imageUtilities.option('max'), 
                        col=c(255,0,0) ) 
{
    if ( is.character(col) ) col <- as.vector(col2rgb(col))
    if ( !is.vector(col) ) col <- as.vector(col)

    ovl <- pmin( ovl, 1 )
    nImg = ifelse(length(dim(img))==2,1,dim(img)[3])
    nOvl = ifelse(length(dim(ovl))==2,1,dim(ovl)[3])
    nCol = length(col)
    n    = base::max(nImg,nOvl,nCol)

    outI <- array(img,dim=c(dim(img)[1:2],n))
    if ( length(dim(ovl))==2 )
        dim(ovl)<-c(dim(ovl),1)

    for ( ch in 1:n )
        outI[,,ch] <- transparency*min(col[min(ch,nCol)],max)*
                                   ovl[,,min(ch,nOvl)] +
                      (1-transparency)*img[,,min(ch,nImg)]
    
    return(outI)
}

# @TODO: "no visible global binding" ??
# undocumented
    # NOTE: colorRamp has 255 hard coded in. But we can test!
#' @keywords internal
.COLOUR.RAMP.MAX <- max(colorRamp(c('white','white'))(1))

#' colourise an image according to a custom colour scale, returning
#' calculated image.
#' @param img single-channel input image to colourise
#' @param colours vector of colour names you want the scale to have, in
#' ascending order.
#' @param bias The colours will be raised to the power of `bias`
#'            before being scaled.
#'            Default 1 (linear color ramp).
#'            Choosing `bias`<1 gives more widely spaced colours at the low end.
#'            Choosing `bias`>1 gives more widely spaced colours at the high end.
#' @param max the max intensity of the input image (will be used for 
#'            output image too).
#' @return image of same dimensions as input except 3 channels. The
#' intensities range from 0 to \code{max}.
#' @note this really will only work for a single-channel input image.
#' @examples
#' \dontrun{
#' img <- array(runif(100),dim=c(10,10))
#' # see the grayscale image
#' drawImage(img,max=1)
#' # apply a new map.
#' colourised <- imColour(img,c('black','red','orange','yellow','white'),
#'                        max=1)
#' drawImage(colourised,max=1)
#' }
#' @export
imColour <- function( img, colours, bias=1,
                     max=imageUtilities.option('max') ) 
{
    ramp <- colorRamp(colours,bias)
    pal  <- ramp(seq.int(0,1,length.out=256))
    if ( max != .COLOUR.RAMP.MAX )
        pal = pal / .COLOUR.RAMP.MAX * max

    nx <- imgetnx(img,1)
    ny <- imgetny(img,1)
    # need to get idxI to go from 0 to 255 since length.out=256.
    # +1 because we're using it as an index into an R object, 1-based
    idxI <- round(as.array(img)*255)/max +1 

    IcolA <- pal[ idxI, ]
    dim(IcolA) <- c( ny,nx,3 )
    return( array2im(IcolA) )
}

