#' @include common_imaging_utilities.r
NULL

#' Overlay a binary image onto another image, displaying it using grid.
#'
#' @param img  the image to overlay onto
#' @param ovl  the *binary* image to be overlaid onto img.
#' @param windowTitle The title to use for the overlay (default being the
#'                     variable name of the overlay image...)
#'                    *NOT USED*
# @param openSaveDialog boolean, default FALSE, whether to open up a "save
#                       as" dialogue after showing the overlay.
#' @param ... ignored.
#' @return the connection to the viewport
#' @note This just displays the overlay on top of the image. If you want
#'  the actual image object consisting of the overlay on the image, use
#'  \code{\link{makeOverlay}}.
#' @family overlay
#' @export
showOverlay <- function(img,ovl,windowTitle=deparse(substitute(ovl)),...)
{
    ovl <- makeOverlay(img,ovl)
    return(drawImage(ovl))
}

#' overlays a single-channel binary image onto another.
#'
#' @inheritParams makeOverlay.grid
#' @return an image with the overlay on top of it in the specified
#'         transparency and colour.
#'
#' @note This essentially performs the calcuation:
#' \deqn{ outI = transparency * colour * overlay + 
#'               (1-transparency) * image }
#' @family overlay
#' @export
makeOverlay <- makeOverlay.grid

#' displays an image, alias for \code{\link{drawImage}}
#' @param img image to be displayed
#' @param fractionOfScreen fraction of screen to take up (may or may not be
#' honoured...)
#' @param ... passed into \code{\link{drawImage}}
#' @return viewport the image is drawn on to.
#' @family draw
#' @export
imdisp <- function(img, fractionOfScreen=0.5, ...)
{
    return(drawImage(img,
                     maxWidth=unit(fractionOfScreen,'npc'),
                     ...))
}
